@@ -1,14 +1,24 @@
+import importlib
+
 import pytest
 from fastapi.testclient import TestClient
 
-from ...utils import needs_pydanticv2
+from ...utils import needs_py39, needs_pydanticv2
 
 
-@pytest.fixture(name="client")
-def get_client():
-    from docs_src.path_operation_advanced_configuration.tutorial007 import app
+@pytest.fixture(
+    name="client",
+    params=[
+        pytest.param("tutorial007"),
+        pytest.param("tutorial007_py39", marks=needs_py39),
+    ],
+)
+def get_client(request: pytest.FixtureRequest):
+    mod = importlib.import_module(
+        f"docs_src.path_operation_advanced_configuration.{request.param}"
+    )
 
-    client = TestClient(app)
+    client = TestClient(mod.app)
     return client
 
 
